unit threadedimageLoader;

//22.6.2010 Theo

{$MODE objfpc}{$H+}

interface

uses
  Classes, SysUtils, contnrs, types, syncobjs, Forms, Graphics, fileutil, FPimage;

type

  EThreadedImageLoaderError = class(Exception);
  TLoadState = (lsEmpty, lsLoading, lsLoaded, lsError);

  { TThumbImage }

  TThumbImage = class
  private
    FArea: TRect;
    FBitmap: TBitmap;
    FHeight: integer;
    FImage: TFPMemoryImage;
    FOnLoaded: TNotifyEvent;
    FOnLoadURL: TNotifyEvent;
    FOnLoadStr: TNotifyEvent;
    FStream: TStream;
    FURL: UTF8String;
    FLoadState: TLoadState;
    FUserData:TObject;
    FRect: TRect;
    FWidth: integer;
    FMultiThreaded: Boolean;
    function GetImage: TFPMemoryImage;
    function GetRect: TRect;
    procedure ThreadTerm(Sender: TObject);

  public
    constructor Create; overload;
    constructor Create(URL: UTF8String); overload;
    constructor Create(URL: UTF8String;str:TStream); overload;
    destructor Destroy; override;
    function Load(Reload: Boolean = False): Boolean;
    procedure FreeImage;
    procedure setUserData(data:TOBject);
    function getUserData():TOBject;
    property Image: TFPMemoryImage read GetImage;
    property Bitmap: TBitmap read FBitmap write FBitmap;
    property LoadState: TLoadState read FLoadState write FLoadState;
    property URL: UTF8String read FURL write FURL;
    property Stream: TStream read FStream write FStream;
    property Left: integer read FRect.Left write FRect.Left;
    property Top: integer read FRect.Top write FRect.Top;
    property Width: integer read FWidth write FWidth;
    property Height: integer read FHeight write FHeight;
    property Rect: TRect read GetRect;
    property Area: TRect read fArea write fArea;
  published
    property OnLoaded: TNotifyEvent read FOnLoaded write FOnLoaded;
    property OnLoadURL: TNotifyEvent read FOnLoadURL write FOnLoadURL;
    property OnLoadStr: TNotifyEvent read FOnLoadStr write FOnLoadStr;
  end;


  { TImageLoaderManager }

  TImageLoaderManager = class
  private
    FActiveIndex: integer;
    FFreeInvisibleImage: Boolean;
    fList: TObjectList;
    FQueue: TList;
    FOnLoaded: TNotifyEvent;
    FOnLoadURL: TNotifyEvent;
    FOnLoadStr: TNotifyEvent;
    FReload: Boolean;
    function GetActiveItem: TThumbImage;
    procedure SetActiveIndex(const AValue: integer);
  public
    constructor Create;
    destructor Destroy; override;
    function AddImage(URL: UTF8String): TThumbImage;
    function AddStreamImage(URL: UTF8String;str:TStream): TThumbImage;
    procedure LoadAll;
    procedure LoadRect(ARect: TRect);
    function ItemIndexFromPoint(Point: TPoint): integer;
    function ItemFromPoint(Point: TPoint): TThumbImage;
    procedure Clear;
    procedure FreeImages;
    function CountItems: integer;
    property List: TObjectList read fList write fList;
    property Reload: Boolean read FReload write FReload;
    property FreeInvisibleImage: Boolean read FFreeInvisibleImage write FFreeInvisibleImage;
    procedure Sort(stpye: byte);
    function ItemFromIndex(AValue: integer): TThumbImage;
    property ActiveIndex: integer read FActiveIndex write SetActiveIndex;
    property ActiveItem: TThumbImage read GetActiveItem;

  published
    property OnLoaded: TNotifyEvent read FOnLoaded write FOnLoaded;
    property OnLoadURL: TNotifyEvent read FOnLoadURL write FOnLoadURL;
    property OnLoadStr: TNotifyEvent read FOnLoadStr write FOnLoadStr;

  end;

var CSImg: TCriticalSection;

implementation

uses Math;

var CS: TCriticalSection;


{ TThumbImage }

procedure TThumbImage.ThreadTerm(Sender: TObject);
var aW, aH: integer;
begin
  FLoadState := lsEmpty;
  if FImage <> nil then
  begin
    aW := fImage.Width;
    aH := fImage.Height;
    if (aW > 0) and (aH > 0) then
    begin
      if fBitmap = nil then fBitmap := TBitmap.Create;
      try
        try
          CSImg.Acquire;
          fBitmap.Assign(fImage);
          fBitmap.Transparent:=false;
          FLoadState := lsError;
          FreeAndNil(fImage);
        except
        end;
      finally
        CSImg.Release;
      end;
      FLoadState := lsLoaded;
      if fMultiThreaded then if Assigned(fOnLoaded) then OnLoaded(Self);
    end;
  end;

end;

function TThumbImage.GetRect: TRect;
begin
  fRect.Right := fRect.Left + fWidth;
  fRect.Bottom := fRect.Top + fHeight;
  Result := fRect;
end;


function TThumbImage.GetImage: TFPMemoryImage;
begin
  Result := FImage;
end;

constructor TThumbImage.Create;
begin
  FLoadState := lsEmpty;
  fMultiThreaded := False;
  FBitmap := nil;
  Fimage := nil;
  FUserData:=nil;
end;

constructor TThumbImage.Create(URL: UTF8String);
begin
  fURL := URL;
  Create;
end;
constructor TThumbImage.Create(URL: UTF8String;str:TStream); overload;
begin
  fURL := URL;
  FStream:=str;
  Create;
end;

destructor TThumbImage.Destroy;
begin
  fBitmap.free;
  inherited Destroy;
end;

function TThumbImage.Load(Reload: Boolean): Boolean;
begin
  Result := True;
  if URL = '' then begin Result := False; exit; end;

  if (fLoadState = lsEmpty) or Reload then
  begin
    fLoadState := lsLoading;
    if Fimage = nil then
    begin
      FImage := TFPMemoryImage.Create(0, 0);
      FImage.UsePalette := false;
    end;


         if assigned(FStream) then
          begin
          if Assigned(FOnLoadStr) then OnLoadStr(Self) else Image.LoadFromStream(Stream);
          end else
          begin
          if Assigned(fOnLoadURL) then OnLoadURL(Self) else Image.LoadFromFile(URL);
          end;
      ThreadTerm(self);

  end else Result := false;
end;
procedure TThumbImage.setUserData(data:TOBject);
begin
FUserData:=data;
end;

function TThumbImage.getUserData():TOBject;
begin
result:=FUserData;
end;

procedure TThumbImage.FreeImage;
begin
  fLoadState := lsEmpty;
  FreeAndNil(fBitmap);
end;





{ TImageLoaderManager }


function TImageLoaderManager.GetActiveItem: TThumbImage;
begin
  Result := ItemFromIndex(fActiveIndex);
end;

procedure TImageLoaderManager.SetActiveIndex(const AValue: integer);
begin
  if (AValue > -1) and (AValue < fList.Count) then FActiveIndex := AValue;
end;

constructor TImageLoaderManager.Create;
begin

  FList := TObjectList.Create;
  FQueue := TList.Create;
  FReload := false;
  FFreeInvisibleImage := false;

end;

destructor TImageLoaderManager.Destroy;
begin
  FQueue.free;
  FList.free;
  inherited Destroy;
end;

function TImageLoaderManager.AddImage(URL: UTF8String): TThumbImage;
begin
  Result := TThumbImage.Create(URL);
  if Assigned(FOnLoadURL) then Result.OnLoadURL := FOnLoadURL;
  if Assigned(FOnLoaded) then Result.OnLoaded := FOnLoaded;
  FList.Add(Result);
end;
function TImageLoaderManager.AddStreamImage(URL: UTF8String;str:TStream): TThumbImage;
begin
  Result := TThumbImage.Create(URL,str);
  if Assigned(FOnLoadStr) then Result.OnLoadStr := FOnLoadStr;
  if Assigned(FOnLoaded) then Result.OnLoaded := FOnLoaded;
  FList.Add(Result);
end;

procedure TImageLoaderManager.LoadAll;
var i: integer;
begin

    for i := 0 to fList.Count - 1 do TThumbImage(fList[i]).Load(FReload) ;

end;

procedure TImageLoaderManager.LoadRect(ARect: TRect);
var i: integer;
  Dum: TRect;
begin
  FQueue.Clear;

    for i := 0 to fList.Count - 1 do
      if IntersectRect(Dum, ARect, TThumbImage(fList[i]).Rect) then
        TThumbImage(fList[i]).Load(FReload) else if FFreeInvisibleImage then
        TThumbImage(fList[i]).FreeImage;

end;




function TImageLoaderManager.ItemIndexFromPoint(Point: TPoint): integer;
var i: integer;
  aRect: TRect;
begin
  Result := -1;
  for i := 0 to fList.Count - 1 do
    if TThumbImage(fList[i]).LoadState = lsLoaded then
    begin
      aRect := TThumbImage(fList[i]).Rect;
      OffsetRect(aRect, TThumbImage(fList[i]).Area.Left, TThumbImage(fList[i]).Area.Top);
      if PtInRect(aRect, Point) then
      begin
        Result := i;
        break;
      end;
    end;
end;

function TImageLoaderManager.ItemFromPoint(Point: TPoint): TThumbImage;
begin
  Result := ItemFromIndex(ItemIndexFromPoint(Point));
end;


procedure TImageLoaderManager.Clear;
begin
  FList.Clear;
end;

procedure TImageLoaderManager.FreeImages;
var i: integer;
begin
  try
    CSImg.Acquire;
    for i := 0 to fList.Count - 1 do TThumbImage(fList[i]).FreeImage;
  finally
    CSImg.Release;
  end;
end;

function TImageLoaderManager.CountItems: integer;
begin
  CS.Acquire;
  try
    Result := fList.Count;
  finally
    CS.Release;
  end;
end;


function sComp(Item1, Item2: Pointer): Integer;
begin
  Result := CompareStr(TThumbImage(Item1).URL, TThumbImage(Item2).URL);
end;

procedure TImageLoaderManager.Sort(stpye: byte);
begin
  fList.Sort(@sComp);
end;

function TImageLoaderManager.ItemFromIndex(AValue: integer): TThumbImage;
begin
  if (AValue > -1) and (AValue < fList.Count) then Result := TThumbImage(fList[AValue]) else
    Result := nil;
end;

initialization

  CS := TCriticalSection.Create;
  CSImg := TCriticalSection.Create;

finalization

  CSImg.free;
  CS.Free;


end.
